@@ -2,29 +2,26 @@
 HTML Widget classes
 """
 
-from __future__ import unicode_literals
-
 import copy
 import datetime
 import re
+import warnings
+from collections import defaultdict
 from itertools import chain
 
 from django.conf import settings
 from django.forms.utils import to_current_timezone
 from django.templatetags.static import static
-from django.utils import datetime_safe, formats, six
+from django.utils import datetime_safe, formats
+from django.utils.datastructures import OrderedSet
 from django.utils.dates import MONTHS
-from django.utils.deprecation import (
-    RemovedInDjango20Warning, RenameMethodsBase,
-)
-from django.utils.encoding import (
-    force_str, force_text, python_2_unicode_compatible,
-)
 from django.utils.formats import get_format
 from django.utils.html import format_html, html_safe
 from django.utils.safestring import mark_safe
-from django.utils.six.moves import range
-from django.utils.translation import ugettext_lazy
+from django.utils.topological_sort import (
+    CyclicDependencyError, stable_topological_sort,
+)
+from django.utils.translation import gettext_lazy as _
 
 from .renderers import get_default_renderer
 
@@ -41,26 +38,44 @@ __all__ = (
 MEDIA_TYPES = ('css', 'js')
 
 
+class MediaOrderConflictWarning(RuntimeWarning):
+    pass
+
+
 @html_safe
-@python_2_unicode_compatible
-class Media(object):
-    def __init__(self, media=None, **kwargs):
-        if media:
-            media_attrs = media.__dict__
+class Media:
+    def __init__(self, media=None, css=None, js=None):
+        if media is not None:
+            css = getattr(media, 'css', {})
+            js = getattr(media, 'js', [])
         else:
-            media_attrs = kwargs
+            if css is None:
+                css = {}
+            if js is None:
+                js = []
+        self._css_lists = [css]
+        self._js_lists = [js]
 
-        self._css = {}
-        self._js = []
-
-        for name in MEDIA_TYPES:
-            getattr(self, 'add_' + name)(media_attrs.get(name))
+    def __repr__(self):
+        return 'Media(css=%r, js=%r)' % (self._css, self._js)
 
     def __str__(self):
         return self.render()
 
+    @property
+    def _css(self):
+        css = defaultdict(list)
+        for css_list in self._css_lists:
+            for medium, sublist in css_list.items():
+                css[medium].append(sublist)
+        return {medium: self.merge(*lists) for medium, lists in css.items()}
+
+    @property
+    def _js(self):
+        return self.merge(*self._js_lists)
+
     def render(self):
-        return mark_safe('\n'.join(chain(*[getattr(self, 'render_' + name)() for name in MEDIA_TYPES])))
+        return mark_safe('\n'.join(chain.from_iterable(getattr(self, 'render_' + name)() for name in MEDIA_TYPES)))
 
     def render_js(self):
         return [
@@ -73,13 +88,13 @@ class Media(object):
     def render_css(self):
         # To keep rendering order consistent, we can't just iterate over items().
         # We need to sort the keys, and iterate over the sorted list.
-        media = sorted(self._css.keys())
-        return chain(*[[
+        media = sorted(self._css)
+        return chain.from_iterable([
             format_html(
-                '<link href="{}" type="text/css" media="{}" rel="stylesheet" />',
+                '<link href="{}" type="text/css" media="{}" rel="stylesheet">',
                 self.absolute_path(path), medium
             ) for path in self._css[medium]
-        ] for medium in media])
+        ] for medium in media)
 
     def absolute_path(self, path):
         """
@@ -92,29 +107,48 @@ class Media(object):
         return static(path)
 
     def __getitem__(self, name):
-        "Returns a Media object that only contains media of the given type"
+        """Return a Media object that only contains media of the given type."""
         if name in MEDIA_TYPES:
             return Media(**{str(name): getattr(self, '_' + name)})
         raise KeyError('Unknown media type "%s"' % name)
 
-    def add_js(self, data):
-        if data:
-            for path in data:
-                if path not in self._js:
-                    self._js.append(path)
+    @staticmethod
+    def merge(*lists):
+        """
+        Merge lists while trying to keep the relative order of the elements.
+        Warn if the lists have the same elements in a different relative order.
 
-    def add_css(self, data):
-        if data:
-            for medium, paths in data.items():
-                for path in paths:
-                    if not self._css.get(medium) or path not in self._css[medium]:
-                        self._css.setdefault(medium, []).append(path)
+        For static assets it can be important to have them included in the DOM
+        in a certain order. In JavaScript you may not be able to reference a
+        global or in CSS you might want to override a style.
+        """
+        dependency_graph = defaultdict(set)
+        all_items = OrderedSet()
+        for list_ in filter(None, lists):
+            head = list_[0]
+            # The first items depend on nothing but have to be part of the
+            # dependency graph to be included in the result.
+            dependency_graph.setdefault(head, set())
+            for item in list_:
+                all_items.add(item)
+                # No self dependencies
+                if head != item:
+                    dependency_graph[item].add(head)
+                head = item
+        try:
+            return stable_topological_sort(all_items, dependency_graph)
+        except CyclicDependencyError:
+            warnings.warn(
+                'Detected duplicate Media files in an opposite order: {}'.format(
+                    ', '.join(repr(l) for l in lists)
+                ), MediaOrderConflictWarning,
+            )
+            return list(all_items)
 
     def __add__(self, other):
         combined = Media()
-        for name in MEDIA_TYPES:
-            getattr(combined, 'add_' + name)(getattr(self, '_' + name, None))
-            getattr(combined, 'add_' + name)(getattr(other, '_' + name, None))
+        combined._css_lists = self._css_lists + other._css_lists
+        combined._js_lists = self._js_lists + other._js_lists
         return combined
 
 
@@ -139,10 +173,8 @@ def media_property(cls):
                     for medium in extend:
                         m = m + base[medium]
                 return m + Media(definition)
-            else:
-                return Media(definition)
-        else:
-            return base
+            return Media(definition)
+        return base
     return property(_media)
 
 
@@ -159,23 +191,14 @@ class MediaDefiningClass(type):
         return new_class
 
 
-class RenameWidgetMethods(MediaDefiningClass, RenameMethodsBase):
-    renamed_methods = (
-        ('_format_value', 'format_value', RemovedInDjango20Warning),
-    )
-
-
-class Widget(six.with_metaclass(RenameWidgetMethods)):
+class Widget(metaclass=MediaDefiningClass):
     needs_multipart_form = False  # Determines does this widget need multipart form
     is_localized = False
     is_required = False
     supports_microseconds = True
 
     def __init__(self, attrs=None):
-        if attrs is not None:
-            self.attrs = attrs.copy()
-        else:
-            self.attrs = {}
+        self.attrs = {} if attrs is None else attrs.copy()
 
     def __deepcopy__(self, memo):
         obj = copy.copy(self)
@@ -199,7 +222,7 @@ class Widget(six.with_metaclass(RenameWidgetMethods)):
             return None
         if self.is_localized:
             return formats.localize_input(value)
-        return force_text(value)
+        return str(value)
 
     def get_context(self, name, value, attrs):
         context = {}
@@ -214,9 +237,7 @@ class Widget(six.with_metaclass(RenameWidgetMethods)):
         return context
 
     def render(self, name, value, attrs=None, renderer=None):
-        """
-        Returns this Widget rendered as HTML, as a Unicode string.
-        """
+        """Render the widget as an HTML string."""
         context = self.get_context(name, value, attrs)
         return self._render(self.template_name, context, renderer)
 
@@ -226,16 +247,13 @@ class Widget(six.with_metaclass(RenameWidgetMethods)):
         return mark_safe(renderer.render(template_name, context))
 
     def build_attrs(self, base_attrs, extra_attrs=None):
-        "Helper function for building an attribute dictionary."
-        attrs = base_attrs.copy()
-        if extra_attrs is not None:
-            attrs.update(extra_attrs)
-        return attrs
+        """Build an attribute dictionary."""
+        return {**base_attrs, **(extra_attrs or {})}
 
     def value_from_datadict(self, data, files, name):
         """
-        Given a dictionary of data and this widget's name, returns the value
-        of this widget. Returns None if it's not provided.
+        Given a dictionary of data and this widget's name, return the value
+        of this widget or None if it's not provided.
         """
         return data.get(name)
 
@@ -244,8 +262,8 @@ class Widget(six.with_metaclass(RenameWidgetMethods)):
 
     def id_for_label(self, id_):
         """
-        Returns the HTML ID attribute of this Widget for use by a <label>,
-        given the ID of the field. Returns None if no ID is available.
+        Return the HTML ID attribute of this Widget for use by a <label>,
+        given the ID of the field. Return None if no ID is available.
 
         This hook is necessary because some widgets have multiple HTML
         elements and, thus, multiple IDs. In that case, this method should
@@ -269,10 +287,10 @@ class Input(Widget):
         if attrs is not None:
             attrs = attrs.copy()
             self.input_type = attrs.pop('type', self.input_type)
-        super(Input, self).__init__(attrs)
+        super().__init__(attrs)
 
     def get_context(self, name, value, attrs):
-        context = super(Input, self).get_context(name, value, attrs)
+        context = super().get_context(name, value, attrs)
         context['widget']['type'] = self.input_type
         return context
 
@@ -302,13 +320,13 @@ class PasswordInput(Input):
     template_name = 'django/forms/widgets/password.html'
 
     def __init__(self, attrs=None, render_value=False):
-        super(PasswordInput, self).__init__(attrs)
+        super().__init__(attrs)
         self.render_value = render_value
 
     def get_context(self, name, value, attrs):
         if not self.render_value:
             value = None
-        return super(PasswordInput, self).get_context(name, value, attrs)
+        return super().get_context(name, value, attrs)
 
 
 class HiddenInput(Input):
@@ -318,13 +336,13 @@ class HiddenInput(Input):
 
 class MultipleHiddenInput(HiddenInput):
     """
-    A widget that handles <input type="hidden"> for fields that have a list
+    Handle <input type="hidden"> for fields that have a list
     of values.
     """
     template_name = 'django/forms/widgets/multiple_hidden.html'
 
     def get_context(self, name, value, attrs):
-        context = super(MultipleHiddenInput, self).get_context(name, value, attrs)
+        context = super().get_context(name, value, attrs)
         final_attrs = context['widget']['attrs']
         id_ = context['widget']['attrs'].get('id')
 
@@ -374,9 +392,9 @@ FILE_INPUT_CONTRADICTION = object()
 
 
 class ClearableFileInput(FileInput):
-    clear_checkbox_label = ugettext_lazy('Clear')
-    initial_text = ugettext_lazy('Currently')
-    input_text = ugettext_lazy('Change')
+    clear_checkbox_label = _('Clear')
+    initial_text = _('Currently')
+    input_text = _('Change')
     template_name = 'django/forms/widgets/clearable_file_input.html'
 
     def clear_checkbox_name(self, name):
@@ -406,7 +424,7 @@ class ClearableFileInput(FileInput):
             return value
 
     def get_context(self, name, value, attrs):
-        context = super(ClearableFileInput, self).get_context(name, value, attrs)
+        context = super().get_context(name, value, attrs)
         checkbox_name = self.clear_checkbox_name(name)
         checkbox_id = self.clear_checkbox_id(checkbox_name)
         context['widget'].update({
@@ -420,7 +438,7 @@ class ClearableFileInput(FileInput):
         return context
 
     def value_from_datadict(self, data, files, name):
-        upload = super(ClearableFileInput, self).value_from_datadict(data, files, name)
+        upload = super().value_from_datadict(data, files, name)
         if not self.is_required and CheckboxInput().value_from_datadict(
                 data, files, self.clear_checkbox_name(name)):
 
@@ -434,11 +452,11 @@ class ClearableFileInput(FileInput):
         return upload
 
     def use_required_attribute(self, initial):
-        return super(ClearableFileInput, self).use_required_attribute(initial) and not initial
+        return super().use_required_attribute(initial) and not initial
 
     def value_omitted_from_data(self, data, files, name):
         return (
-            super(ClearableFileInput, self).value_omitted_from_data(data, files, name) and
+            super().value_omitted_from_data(data, files, name) and
             self.clear_checkbox_name(name) not in data
         )
 
@@ -451,7 +469,7 @@ class Textarea(Widget):
         default_attrs = {'cols': '40', 'rows': '10'}
         if attrs:
             default_attrs.update(attrs)
-        super(Textarea, self).__init__(default_attrs)
+        super().__init__(default_attrs)
 
 
 class DateTimeBaseInput(TextInput):
@@ -459,13 +477,11 @@ class DateTimeBaseInput(TextInput):
     supports_microseconds = False
 
     def __init__(self, attrs=None, format=None):
-        super(DateTimeBaseInput, self).__init__(attrs)
-        self.format = format if format else None
+        super().__init__(attrs)
+        self.format = format or None
 
     def format_value(self, value):
-        if value is not None:
-            # localize_input() returns str on Python 2.
-            return force_text(formats.localize_input(value, self.format or formats.get_format(self.format_key)[0]))
+        return formats.localize_input(value, self.format or formats.get_format(self.format_key)[0])
 
 
 class DateInput(DateTimeBaseInput):
@@ -493,7 +509,7 @@ class CheckboxInput(Input):
     template_name = 'django/forms/widgets/checkbox.html'
 
     def __init__(self, attrs=None, check_test=None):
-        super(CheckboxInput, self).__init__(attrs)
+        super().__init__(attrs)
         # check_test is a callable that takes a value and returns True
         # if the checkbox should be checked for that value.
         self.check_test = boolean_check if check_test is None else check_test
@@ -502,14 +518,14 @@ class CheckboxInput(Input):
         """Only return the 'value' attribute if value isn't empty."""
         if value is True or value is False or value is None or value == '':
             return
-        return force_text(value)
+        return str(value)
 
     def get_context(self, name, value, attrs):
         if self.check_test(value):
             if attrs is None:
                 attrs = {}
             attrs['checked'] = True
-        return super(CheckboxInput, self).get_context(name, value, attrs)
+        return super().get_context(name, value, attrs)
 
     def value_from_datadict(self, data, files, name):
         if name not in data:
@@ -519,7 +535,7 @@ class CheckboxInput(Input):
         value = data.get(name)
         # Translate true and false strings to boolean values.
         values = {'true': True, 'false': False}
-        if isinstance(value, six.string_types):
+        if isinstance(value, str):
             value = values.get(value.lower(), value)
         return bool(value)
 
@@ -539,7 +555,7 @@ class ChoiceWidget(Widget):
     option_inherits_attrs = True
 
     def __init__(self, attrs=None, choices=()):
-        super(ChoiceWidget, self).__init__(attrs)
+        super().__init__(attrs)
         # choices can be any iterable, but we may need to render this widget
         # multiple times. Thus, collapse it into a list so it can be consumed
         # more than once.
@@ -558,21 +574,19 @@ class ChoiceWidget(Widget):
         options from a BoundField for choice widgets.
         """
         value = self.format_value(value)
-        for option in self.options(name, value, attrs):
-            yield option
+        yield from self.options(name, value, attrs)
 
     def options(self, name, value, attrs=None):
         """Yield a flat list of options for this widgets."""
         for group in self.optgroups(name, value, attrs):
-            for option in group[1]:
-                yield option
+            yield from group[1]
 
     def optgroups(self, name, value, attrs=None):
         """Return a list of optgroups for this widget."""
         groups = []
         has_selected = False
 
-        for index, (option_value, option_label) in enumerate(chain(self.choices)):
+        for index, (option_value, option_label) in enumerate(self.choices):
             if option_value is None:
                 option_value = ''
 
@@ -589,11 +603,10 @@ class ChoiceWidget(Widget):
 
             for subvalue, sublabel in choices:
                 selected = (
-                    force_text(subvalue) in value and
-                    (has_selected is False or self.allow_multiple_selected)
+                    str(subvalue) in value and
+                    (not has_selected or self.allow_multiple_selected)
                 )
-                if selected is True and has_selected is False:
-                    has_selected = True
+                has_selected |= selected
                 subgroup.append(self.create_option(
                     name, subvalue, sublabel, selected, index,
                     subindex=subindex, attrs=attrs,
@@ -620,12 +633,12 @@ class ChoiceWidget(Widget):
             'attrs': option_attrs,
             'type': self.input_type,
             'template_name': self.option_template_name,
+            'wrap_label': True,
         }
 
     def get_context(self, name, value, attrs):
-        context = super(ChoiceWidget, self).get_context(name, value, attrs)
+        context = super().get_context(name, value, attrs)
         context['widget']['optgroups'] = self.optgroups(name, context['widget']['value'], attrs)
-        context['wrap_label'] = True
         return context
 
     def id_for_label(self, id_, index='0'):
@@ -648,9 +661,11 @@ class ChoiceWidget(Widget):
 
     def format_value(self, value):
         """Return selected values as a list."""
+        if value is None and self.allow_multiple_selected:
+            return []
         if not isinstance(value, (tuple, list)):
             value = [value]
-        return [force_text(v) if v is not None else '' for v in value]
+        return [str(v) if v is not None else '' for v in value]
 
 
 class Select(ChoiceWidget):
@@ -662,26 +677,23 @@ class Select(ChoiceWidget):
     option_inherits_attrs = False
 
     def get_context(self, name, value, attrs):
-        context = super(Select, self).get_context(name, value, attrs)
+        context = super().get_context(name, value, attrs)
         if self.allow_multiple_selected:
-            context['widget']['attrs']['multiple'] = 'multiple'
+            context['widget']['attrs']['multiple'] = True
         return context
 
     @staticmethod
     def _choice_has_empty_value(choice):
         """Return True if the choice's value is empty string or None."""
         value, _ = choice
-        return (
-            (isinstance(value, six.string_types) and not bool(value)) or
-            value is None
-        )
+        return value is None or value == ''
 
     def use_required_attribute(self, initial):
         """
         Don't render 'required' if the first <option> has a value, as that's
         invalid HTML.
         """
-        use_required_attribute = super(Select, self).use_required_attribute(initial)
+        use_required_attribute = super().use_required_attribute(initial)
         # 'required' is always okay for <select multiple>.
         if self.allow_multiple_selected:
             return use_required_attribute
@@ -696,27 +708,35 @@ class NullBooleanSelect(Select):
     """
     def __init__(self, attrs=None):
         choices = (
-            ('1', ugettext_lazy('Unknown')),
-            ('2', ugettext_lazy('Yes')),
-            ('3', ugettext_lazy('No')),
+            ('unknown', _('Unknown')),
+            ('true', _('Yes')),
+            ('false', _('No')),
         )
-        super(NullBooleanSelect, self).__init__(attrs, choices)
+        super().__init__(attrs, choices)
 
     def format_value(self, value):
         try:
-            return {True: '2', False: '3', '2': '2', '3': '3'}[value]
+            return {
+                True: 'true', False: 'false',
+                'true': 'true', 'false': 'false',
+                # For backwards compatibility with Django < 2.2.
+                '2': 'true', '3': 'false',
+            }[value]
         except KeyError:
-            return '1'
+            return 'unknown'
 
     def value_from_datadict(self, data, files, name):
         value = data.get(name)
         return {
-            '2': True,
             True: True,
             'True': True,
-            '3': False,
             'False': False,
             False: False,
+            'true': True,
+            'false': False,
+            # For backwards compatibility with Django < 2.2.
+            '2': True,
+            '3': False,
         }.get(value)
 
 
@@ -765,7 +785,7 @@ class CheckboxSelectMultiple(ChoiceWidget):
         """
         if index is None:
             return ''
-        return super(CheckboxSelectMultiple, self).id_for_label(id_, index)
+        return super().id_for_label(id_, index)
 
 
 class MultiWidget(Widget):
@@ -782,14 +802,14 @@ class MultiWidget(Widget):
 
     def __init__(self, widgets, attrs=None):
         self.widgets = [w() if isinstance(w, type) else w for w in widgets]
-        super(MultiWidget, self).__init__(attrs)
+        super().__init__(attrs)
 
     @property
     def is_hidden(self):
         return all(w.is_hidden for w in self.widgets)
 
     def get_context(self, name, value, attrs):
-        context = super(MultiWidget, self).get_context(name, value, attrs)
+        context = super().get_context(name, value, attrs)
         if self.is_localized:
             for widget in self.widgets:
                 widget.is_localized = self.is_localized
@@ -835,14 +855,17 @@ class MultiWidget(Widget):
 
     def decompress(self, value):
         """
-        Returns a list of decompressed values for the given compressed value.
+        Return a list of decompressed values for the given compressed value.
         The given value can be assumed to be valid, but not necessarily
         non-empty.
         """
         raise NotImplementedError('Subclasses must implement this method.')
 
     def _get_media(self):
-        "Media for a multiwidget is the combination of all media of the subwidgets"
+        """
+        Media for a multiwidget is the combination of all media of the
+        subwidgets.
+        """
         media = Media()
         for w in self.widgets:
             media = media + w.media
@@ -850,7 +873,7 @@ class MultiWidget(Widget):
     media = property(_get_media)
 
     def __deepcopy__(self, memo):
-        obj = super(MultiWidget, self).__deepcopy__(memo)
+        obj = super().__deepcopy__(memo)
         obj.widgets = copy.deepcopy(self.widgets)
         return obj
 
@@ -861,52 +884,58 @@ class MultiWidget(Widget):
 
 class SplitDateTimeWidget(MultiWidget):
     """
-    A Widget that splits datetime input into two <input type="text"> boxes.
+    A widget that splits datetime input into two <input type="text"> boxes.
     """
     supports_microseconds = False
     template_name = 'django/forms/widgets/splitdatetime.html'
 
-    def __init__(self, attrs=None, date_format=None, time_format=None):
+    def __init__(self, attrs=None, date_format=None, time_format=None, date_attrs=None, time_attrs=None):
         widgets = (
-            DateInput(attrs=attrs, format=date_format),
-            TimeInput(attrs=attrs, format=time_format),
+            DateInput(
+                attrs=attrs if date_attrs is None else date_attrs,
+                format=date_format,
+            ),
+            TimeInput(
+                attrs=attrs if time_attrs is None else time_attrs,
+                format=time_format,
+            ),
         )
-        super(SplitDateTimeWidget, self).__init__(widgets, attrs)
+        super().__init__(widgets)
 
     def decompress(self, value):
         if value:
             value = to_current_timezone(value)
-            return [value.date(), value.time().replace(microsecond=0)]
+            return [value.date(), value.time()]
         return [None, None]
 
 
 class SplitHiddenDateTimeWidget(SplitDateTimeWidget):
     """
-    A Widget that splits datetime input into two <input type="hidden"> inputs.
+    A widget that splits datetime input into two <input type="hidden"> inputs.
     """
     template_name = 'django/forms/widgets/splithiddendatetime.html'
 
-    def __init__(self, attrs=None, date_format=None, time_format=None):
-        super(SplitHiddenDateTimeWidget, self).__init__(attrs, date_format, time_format)
+    def __init__(self, attrs=None, date_format=None, time_format=None, date_attrs=None, time_attrs=None):
+        super().__init__(attrs, date_format, time_format, date_attrs, time_attrs)
         for widget in self.widgets:
             widget.input_type = 'hidden'
 
 
 class SelectDateWidget(Widget):
     """
-    A Widget that splits date input into three <select> boxes.
+    A widget that splits date input into three <select> boxes.
 
     This also serves as an example of a Widget that has more than one HTML
     element and hence implements value_from_datadict.
     """
-    none_value = (0, '---')
+    none_value = ('', '---')
     month_field = '%s_month'
     day_field = '%s_day'
     year_field = '%s_year'
     template_name = 'django/forms/widgets/select_date.html'
     input_type = 'select'
     select_widget = Select
-    date_re = re.compile(r'(\d{4})-(\d\d?)-(\d\d?)$')
+    date_re = re.compile(r'(\d{4}|0)-(\d\d?)-(\d\d?)$')
 
     def __init__(self, attrs=None, years=None, months=None, empty_label=None):
         self.attrs = attrs or {}
@@ -929,52 +958,46 @@ class SelectDateWidget(Widget):
             if not len(empty_label) == 3:
                 raise ValueError('empty_label list/tuple must have 3 elements.')
 
-            self.year_none_value = (0, empty_label[0])
-            self.month_none_value = (0, empty_label[1])
-            self.day_none_value = (0, empty_label[2])
+            self.year_none_value = ('', empty_label[0])
+            self.month_none_value = ('', empty_label[1])
+            self.day_none_value = ('', empty_label[2])
         else:
             if empty_label is not None:
-                self.none_value = (0, empty_label)
+                self.none_value = ('', empty_label)
 
             self.year_none_value = self.none_value
             self.month_none_value = self.none_value
             self.day_none_value = self.none_value
 
     def get_context(self, name, value, attrs):
-        context = super(SelectDateWidget, self).get_context(name, value, attrs)
+        context = super().get_context(name, value, attrs)
         date_context = {}
-        year_choices = [(i, i) for i in self.years]
-        if self.is_required is False:
+        year_choices = [(i, str(i)) for i in self.years]
+        if not self.is_required:
             year_choices.insert(0, self.year_none_value)
-        year_attrs = context['widget']['attrs'].copy()
         year_name = self.year_field % name
-        year_attrs['id'] = 'id_%s' % year_name
         date_context['year'] = self.select_widget(attrs, choices=year_choices).get_context(
             name=year_name,
             value=context['widget']['value']['year'],
-            attrs=year_attrs,
+            attrs={**context['widget']['attrs'], 'id': 'id_%s' % year_name},
         )
         month_choices = list(self.months.items())
-        if self.is_required is False:
+        if not self.is_required:
             month_choices.insert(0, self.month_none_value)
-        month_attrs = context['widget']['attrs'].copy()
         month_name = self.month_field % name
-        month_attrs['id'] = 'id_%s' % month_name
         date_context['month'] = self.select_widget(attrs, choices=month_choices).get_context(
             name=month_name,
             value=context['widget']['value']['month'],
-            attrs=month_attrs,
+            attrs={**context['widget']['attrs'], 'id': 'id_%s' % month_name},
         )
         day_choices = [(i, i) for i in range(1, 32)]
-        if self.is_required is False:
+        if not self.is_required:
             day_choices.insert(0, self.day_none_value)
-        day_attrs = context['widget']['attrs'].copy()
         day_name = self.day_field % name
-        day_attrs['id'] = 'id_%s' % day_name
         date_context['day'] = self.select_widget(attrs, choices=day_choices,).get_context(
             name=day_name,
             value=context['widget']['value']['day'],
-            attrs=day_attrs,
+            attrs={**context['widget']['attrs'], 'id': 'id_%s' % day_name},
         )
         subwidgets = []
         for field in self._parse_date_fmt():
@@ -991,17 +1014,20 @@ class SelectDateWidget(Widget):
         year, month, day = None, None, None
         if isinstance(value, (datetime.date, datetime.datetime)):
             year, month, day = value.year, value.month, value.day
-        elif isinstance(value, six.string_types):
-            if settings.USE_L10N:
-                try:
-                    input_format = get_format('DATE_INPUT_FORMATS')[0]
-                    d = datetime.datetime.strptime(force_str(value), input_format)
-                    year, month, day = d.year, d.month, d.day
-                except ValueError:
-                    pass
+        elif isinstance(value, str):
             match = self.date_re.match(value)
             if match:
-                year, month, day = [int(val) for val in match.groups()]
+                # Convert any zeros in the date to empty strings to match the
+                # empty option value.
+                year, month, day = [int(val) or '' for val in match.groups()]
+            elif settings.USE_L10N:
+                input_format = get_format('DATE_INPUT_FORMATS')[0]
+                try:
+                    d = datetime.datetime.strptime(value, input_format)
+                except ValueError:
+                    pass
+                else:
+                    year, month, day = d.year, d.month, d.day
         return {'year': year, 'month': month, 'day': day}
 
     @staticmethod
@@ -1023,27 +1049,27 @@ class SelectDateWidget(Widget):
     def id_for_label(self, id_):
         for first_select in self._parse_date_fmt():
             return '%s_%s' % (id_, first_select)
-        else:
-            return '%s_month' % id_
+        return '%s_month' % id_
 
     def value_from_datadict(self, data, files, name):
         y = data.get(self.year_field % name)
         m = data.get(self.month_field % name)
         d = data.get(self.day_field % name)
-        if y == m == d == "0":
+        if y == m == d == '':
             return None
-        if y and m and d:
+        if y is not None and m is not None and d is not None:
             if settings.USE_L10N:
                 input_format = get_format('DATE_INPUT_FORMATS')[0]
                 try:
                     date_value = datetime.date(int(y), int(m), int(d))
                 except ValueError:
-                    return '%s-%s-%s' % (y, m, d)
+                    pass
                 else:
                     date_value = datetime_safe.new_date(date_value)
                     return date_value.strftime(input_format)
-            else:
-                return '%s-%s-%s' % (y, m, d)
+            # Return pseudo-ISO dates with zeros for any unselected values,
+            # e.g. '2017-0-23'.
+            return '%s-%s-%s' % (y or 0, m or 0, d or 0)
         return data.get(name)
 
     def value_omitted_from_data(self, data, files, name):
